/**********************************************************************
*
* Copyright (c) 2012-2020 Barbara Geller
* Copyright (c) 2012-2020 Ansel Sermersheim
*
* You may use this file under the terms of the 2-Clause BSD license
* provided with PepperMill or available at:
*
* https://opensource.org/licenses/BSD-2-Clause
*
* PepperMill is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*
***********************************************************************/

#include "peppermill.h"

// run arguments
// --onefile z:\PepperMill\filesPM\about.h -f
// --recursive z:\PepperMill\filesPM\cs4_src\corelib\ -f
// --onefolder z:\PepperMill\filesPM\cs4_src\corelib\kernel\ -f
// --onefolder z:\PepperMill\filesPM\cs4_src\gui\kernel\ -f
// --ginger

QByteArray PepperMill::pmParse(const QByteArray &input)
{
   QString output;

   const char *ptr = input.data();
   QString word;

   m_section = Section::NONE;

   m_processFriend = false;
   m_processClass  = false;
   m_consume       = false;

   // pass one
   while (*ptr) {
      char letter = *ptr;

      if (pmIs_WhiteSpace(letter)) {

         if (! word.isEmpty()) {
            // check the token
            word = pmToken(word, &ptr);

            output.append(word);
         }

         output.append(letter);

         // reset
         word = "";

      } else if (letter == ':' || letter == '(' )   {

         if (! word.isEmpty()) {
            // check the token
            word = pmToken(word, &ptr);
            output.append(word);
         }

         if (! m_consume) {
            output.append(letter);
         }

         m_consume = false;

         // reset
         word = "";

      } else if (letter == '#' && pmIs_LineBreak( *(ptr-1) ))  {
         // #ifdef  #ifndef  #define

         while (*ptr) {
            char letter = *ptr;
            word += letter;

            if (letter == '\n' || letter == '\r') {
               break;
            }

            ++ptr;
         }

         output.append(word);

         if (word.contains("define")) {

            while (word.at(1) == ' ' || word.at(1) == '\t')  {
               word.remove(1,1);
            }

            // check the token
            word = pmToken(word, &ptr);
         }

         // reset
         word = "";

         if (! (*ptr)) {
            --ptr;
         }


      } else if (word == "/" && letter == '/')   {
         // single comment

         while (*ptr) {
            char letter = *ptr;
            word += letter;

            if (letter == '\n' || letter == '\r') {
               break;
            }

            ++ptr;
         }

         output.append(word);

         // reset
         word = "";

         if (! (*ptr)) {
            --ptr;
         }

      } else if (word == "/" && letter == '*')   {
         // multi line comment

         while (*ptr) {
            char letter = *ptr;
            word += letter;

            if ( *(ptr-1) == '*' && letter == '/') {
               break;
            }

            ++ptr;
         }

         output.append(word);

         // reset
         word = "";

         if (! (*ptr)) {
            --ptr;
         }

      } else {
         word += letter;

      }

      ++ptr;
   }

   if (! word.isEmpty())  {
      output.append(word);
   }

   // pass two
   int indexStart = 0;
   int indexEnd   = 0;

   QString phrase = "CS_FLAG(x,";
   int size = phrase.size();

   while (true) {

      indexStart = output.indexOf(phrase, indexStart+1);

      if (indexStart == -1) {
         break;
      }

      indexEnd = output.indexOf(")", indexStart);

      QString flagName = output.mid(indexStart + size, indexEnd - indexStart -size);
      QString enumName = m_flagMap.value(flagName);

      if (! enumName.isEmpty()) {
         output.replace(indexStart + size-2, 1, enumName);
      }
   }

   return output.toUtf8();
}

QString PepperMill::pmToken(const QString &word, const char **ptr)
{
   QString newWord = word;

   if (word.left(1) == "}") {
      // could be at end of class

      if (m_section == Section::SIGNAL_PUBLIC || m_section == Section::SLOT_PUBLIC ||
            m_section == Section::SLOT_PROTECTED ||  m_section == Section::SLOT_PRIVATE) {

         // reset
         m_section  = Section::NONE;
      }
   }


   // ** part 1 - order of everything in this part really matters

   if (word == "friend")  {
      m_processFriend = true;

   } else if (m_processFriend) {
      m_processFriend = false;

   } else if (word == "class")  {
      m_processClass = true;
      m_section      = Section::PRIVATE;

   } else if (m_processClass) {

       if (word == "{" || word == "public" || word == "protected" || word == "private" || word.endsWith(';')) {
         m_className    = m_processWord;
         m_processClass = false;

      } else {
         m_processWord = word;
      }

   } else if (word == "Q_OBJECT")  {
      newWord = "CS_OBJECT(" + m_className + ")";

   } else if (word == "Q_GADGET")  {
      newWord = "CS_GADGET(" + m_className + ")";

   } else if (word == "public")   {

      bool isSlot = pmisToken_Slot(ptr);

      if (isSlot) {
         m_section  = Section::SLOT_PUBLIC;
      } else {
         m_section  = Section::PUBLIC;
      }

   } else if (word == "protected")  {

      bool isSlot = pmisToken_Slot(ptr);

      if (isSlot) {
         m_section  = Section::SLOT_PROTECTED;
      } else {
         m_section  = Section::PROTECTED;
      }

   } else if (word == "private")   {

      bool isSlot = pmisToken_Slot(ptr);

      if (isSlot) {
         m_section  = Section::SLOT_PRIVATE;
      } else {
         m_section  = Section::PRIVATE;
      }


   } else if (word == "signals")   {
      m_section  = Section::SIGNAL_PUBLIC;
      newWord = "public";


   } else if (word == "Q_SIGNALS")  {
      newWord = "public";
      m_section = Section::SIGNAL_PUBLIC;

   } else if (word.left(7) == "#define")  {
      // prep for cs_tag
      QList<QString> defineTest = word.split(' ');

      if (defineTest.size() == 2)  {
         m_defines.insert(defineTest.at(1).trimmed());
      }


   // ** part 2
   } else if (m_section == Section::SIGNAL_PUBLIC)  {

      // comment, new section, closing curly -or- method
      bool isMethod = pmisToken_Method(ptr);

      if (isMethod) {
         // must do this before getting the data!
         QString padding = pmPad(*ptr);

         QString data = pmGetMethod(ptr);
         data = word + " " + data.trimmed();

         if (! data.isEmpty()) {

            QString methodName;
            QList<QString> paramNames;
            QString signature;

            std::tie(methodName, paramNames, signature) = pmGetSignature(data);

            QString s2 = methodName;

            for (auto index = paramNames.begin(); index < paramNames.end() ; ++index) {
               s2 += ",";
               s2 += *index;
            }

            newWord  = "CS_SIGNAL_1(Public, " + signature + ")\n";
            newWord += padding + "CS_SIGNAL_2(" + s2 + ")";
         }
      }

   } else if (m_section == Section::SLOT_PUBLIC)  {

      // comment, new section, closing curly -or- method
      bool isMethod = pmisToken_Method(ptr);

      if (isMethod) {
         // must do this before getting the data!
         QString padding = pmPad(*ptr);

         QString data = pmGetMethod(ptr);
         data = word + " " + data.trimmed();

         if (! data.isEmpty()) {

            // cs_tag
            bool isCSTag = false;

            int k = data.indexOf(' ');
            QString defineTest = word.left(k);

            if (m_defines.contains(defineTest)) {
               // match
               data    = data.mid(k+1);
               isCSTag = true;
            }

            QString methodName;
            QList<QString> paramNames;
            QString signature;

            std::tie(methodName, paramNames, signature) = pmGetSignature(data);

/*          QByteArray s2 = methodName;

            for (auto index = paramNames.begin(); index < paramNames.end() ; ++index) {
               s2 += ",";
               s2 += *index;
            }
*/

            if (isCSTag) {
               // any passed parameters are missing from CS_Tag()
               newWord = "CS_TAG(\"" + defineTest + "\", " + methodName + "())\n" + padding;

            } else {
               newWord = "";

            }

            newWord += "CS_SLOT_1(Public, " + signature + ")\n";
            newWord += padding + "CS_SLOT_2(" +  methodName + ")";
         }
      }

   } else if (m_section == Section::SLOT_PROTECTED)  {

      // comment, new section, closing curly -or-  method
      bool isMethod = pmisToken_Method(ptr);

      if (isMethod) {
         // must do this before getting the data!
         QString padding = pmPad(*ptr);

         QString data = pmGetMethod(ptr);
         data = word + " " + data.trimmed();

         if (! data.isEmpty()) {

            // cs_tag
            bool isCSTag = false;

            int k = data.indexOf(' ');
            QString defineTest = word.left(k);

            if (m_defines.contains(defineTest)) {
               // match
               data    = data.mid(k+1);
               isCSTag = true;
            }

            QString methodName;
            QList<QString> paramNames;
            QString signature;

            std::tie(methodName, paramNames, signature) = pmGetSignature(data);

/*          QByteArray s2 = methodName;

            for (auto index = paramNames.begin(); index < paramNames.end() ; ++index) {
               s2 += ",";
               s2 += *index;
            }
*/

            if (isCSTag) {
               // any passed parameters are missing from CS_Tag()
               newWord = "CS_TAG(\"" + defineTest.toLatin1() + "\", " + methodName + "())\n" + padding;

            } else {
               newWord = "";

            }

            newWord += "CS_SLOT_1(Protected, " + signature + ")\n";
            newWord += padding + "CS_SLOT_2(" +  methodName + ")";
         }
      }

   } else if (m_section == Section::SLOT_PRIVATE)  {

      // comment, new section, closing curly -or-  method
      bool isMethod = pmisToken_Method(ptr);

      if (isMethod) {
         // must do this before getting the data!
         QString padding = pmPad(*ptr);

         QString data = pmGetMethod(ptr);
         data = word + " " + data.trimmed();

         if (! data.isEmpty()) {

            // cs_tag
            bool isCSTag = false;

            int k = data.indexOf(' ');
            QString defineTest = word.left(k);

            if (m_defines.contains(defineTest)) {
               // match
               data    = data.mid(k+1);
               isCSTag = true;
            }

            QString methodName;
            QList<QString> paramNames;
            QString signature;

            std::tie(methodName, paramNames, signature) = pmGetSignature(data);

/*          QByteArray s2 = methodName;

            for (auto index = paramNames.begin(); index < paramNames.end() ; ++index) {
               s2 += ",";
               s2 += *index;
            }
*/

            if (isCSTag) {
               // any passed parameters are missing from CS_Tag()
               newWord = "CS_TAG(\"" + defineTest + "\", " + methodName + "())\n" + padding;

            } else {
               newWord = "";

            }

            newWord += "CS_SLOT_1(Private, " + signature + ")\n";
            newWord += padding + "CS_SLOT_2(" +  methodName + ")";
         }
      }


   // ** part 3
   } else if (word == "Q_ARG")  {
      // ** leave alone


   } else if (word == "Q_CLASSINFO")  {
      newWord = "CS_CLASSINFO";


   } else if (word == "Q_DECLARE_EXTENSION_INTERFACE")  {
      // ** leave alone


   } else if (word == "Q_DECLARE_FLAGS")  {
      QString data = pmGetData(ptr);

      // avoids the first parn from being saved
      m_consume = true;

      // exactly two parameters
      int pos = data.indexOf(",");

      QString flagName = data.left(pos).trimmed();
      QString enumName = data.mid(pos+1).trimmed();

      newWord = "using " + flagName + " = QFlags<" + enumName + ">;";

      // save it
      m_flagMap.insert(flagName, enumName);


   } else if (word == "Q_DECLARE_INTERFACE")  {
     newWord = "CS_DECLARE_INTERFACE";


   } else if (word == "Q_DECLARE_METATYPE")  {
     // ** leave alone


   } else if (word == "Q_DECLARE_OPERATORS_FOR_FLAGS")  {
     // ** leave alone


   } else if (word == "Q_DECLARE_PRIVATE")  {
      // ** leave alone, undocumented


   } else if (word == "Q_DECLARE_TR_FUNCTIONS")  {
      // ** leave alone


   } else if (word == "Q_DECLARE_TYPEINFO")  {
       // ** leave alone


   } else if (word == "Q_DISABLE_COPY")  {
      // ** leave alone


   } else if (word == "enum")  {
      // NOT SURE
      // parse m_regEnum for the enum, if it exists


   } else if (word == "Q_ENUMS")  {
      // must do this before getting the data!
      QString padding = pmPad(*ptr);

      // spaced delimited
      QString data = pmGetData(ptr);

      // avoids the first parn from being saved
      m_consume = true;

      QList<QString> temp = data.split(' ');
      temp.removeAll("");
      newWord = "";

      QString preData = "";
      bool isFirst    = true;

      for(auto index = temp.begin(); index != temp.end(); ++index) {

         // save to register this enum
         m_regEnum.append(*index);

         newWord += preData + "CS_ENUM(" + *index + ")\n";

         if (isFirst) {
            preData = padding;
            isFirst = false;
         }
      }

      // back out the last line feed
      if (! isFirst) {
         newWord.chop(1);
      }

   } else if (word == "Q_EMIT")  {
      // ** leave alone


   } else if (word == "Q_EXPORT_PLUGIN2")  {
      // --> change  1  ( unsure )


   } else if (word == "Q_FLAGS")  {
      // spaced delimited
      QString data = pmGetData(ptr);

      // avoids the first parn from being saved
      m_consume = true;

      newWord = "CS_FLAG(Resolve Flag -->," + data + ")" ;


   } else if (word == "Q_IMPORT_PLUGIN")  {
      // --> change  2  ( unsure )


   } else if (word == "Q_INTERFACES")  {
      // must do this before getting the data!
      QString padding = pmPad(*ptr);

      // spaced delimited
      QString data = pmGetData(ptr);

      // avoids the first parn from being saved
      m_consume = true;

      QList<QString> temp = data.split(' ');
      temp.removeAll("");
      newWord = "";

      QString preData = "";
      bool isFirst = true;

      for(auto index = temp.begin(); index != temp.end(); ++index) {
         newWord += preData + "CS_INTERFACES(" + *index + ")\n";

         if (isFirst) {
            preData = padding;
            isFirst = false;
         }
      }

      // back out the last line feed
      if (! isFirst) {
         newWord.chop(1);
      }

   } else if (word == "Q_INVOKEABLE")  {
      // must do this before getting the data!
      QString padding = pmPad(*ptr);

      QString data = pmGetMethod(ptr);
      data = data.trimmed();

      if (! data.isEmpty()) {

         // cs_tag
         bool isCSTag = false;

         int k = data.indexOf(' ');
         QString defineTest = word.left(k);

         if (m_defines.contains(defineTest)) {
            // match
            data    = data.mid(k+1);
            isCSTag = true;
         }

         QString methodName;
         QList<QString> paramNames;
         QString signature;

         std::tie(methodName, paramNames, signature) = pmGetSignature(data);

         QString s2 = methodName;

         for (auto index = paramNames.begin(); index < paramNames.end() ; ++index) {
            s2 += ",";
            s2 += *index;
         }

         QString area;

         if (m_section == Section::PUBLIC)  {
            area = "Public";

         } else if (m_section == Section::PROTECTED)  {
            area = "Protected";

         } else if (m_section == Section::PRIVATE)  {
            area = "Private";

         }

         if (isCSTag) {
            // any passed parameters are missing from CS_Tag()
            newWord = "CS_TAG(\"" + defineTest.toLatin1() + "\", " + methodName + "())\n" + padding;

         } else {
            newWord = "";

         }

         newWord += "CS_INVOKABLE_METHOD_1(" + area + ", " + signature + ")\n";
         newWord += padding + "CS_INVOKABLE_METHOD_2(" + s2 + ")";
      }

   } else if (word == "Q_PRIVATE_PROPERTY")  {
      // ** leave alone, undocumented

   } else if (word == "Q_PRIVATE_SLOT")  {

      // must do this before getting the data!
      QString padding = pmPad(*ptr);

      QString data = pmGetData(ptr);

      // avoids the first parn from being saved
      m_consume = true;

      // exactly two parameters
      int pos = data.indexOf(",");

      QString part2 = data.mid(pos+1).trimmed();
      data = part2;

      if (! data.isEmpty()) {

         // cs_tag
         bool isCSTag = false;

         int k = data.indexOf(' ');
         QString defineTest = word.left(k);

         if (m_defines.contains(defineTest)) {
            // match
            data    = data.mid(k+1);
            isCSTag = true;
         }

         QString methodName;
         QList<QString> paramNames;
         QString signature;

         std::tie(methodName, paramNames, signature) = pmGetSignature(data);

         QString area = "Private";

         if (isCSTag) {
            // any passed parameters are missing from CS_Tag()
            newWord = "CS_TAG(\"" + defineTest.toLatin1() + "\", " + methodName + "())\n" + padding;

         } else {
            newWord = "";

         }

         newWord += "CS_SLOT_1(" + area + ", " + signature + ")\n";
         newWord += padding + "CS_SLOT_2(" +  methodName + ")";

         // write method
         newWord += "\n\n";

         newWord += "/*  PRIVATE_SLOT";
         newWord +=  "\n"    + signature + "\n{\n";
         newWord += "\tQ_D(" + m_className + ");\n";
         newWord += "\td->"  + methodName + "();\n}\n*/";
      }

   } else if (word == "Q_PROPERTY")  {
      // must do this before getting the data!
      QString padding = pmPad(*ptr);

      // spaced delimited
      QString data    = pmGetData(ptr);

      // avoids the first parn from being saved
      m_consume = true;

      QList<QString> temp = data.split(' ');
      temp.removeAll("");

      newWord = "";

      // property variable name
      QString varName = *(temp.begin()+1);

      QString varId;
      QString varFunc;

      QString preData = "";
      bool isFirst = true;

      for(auto index = temp.begin()+2; index != temp.end(); ++index) {

         if (varId.isEmpty()) {
            QString tempId = *index;
            tempId = tempId.trimmed();

            if (tempId == "READ"       || tempId == "WRITE"    || tempId == "RESET" ||
                tempId == "NOTIFY"     || tempId == "REVISION" || tempId == "DESIGNABLE" ||
                tempId == "SCRIPTABLE" || tempId == "STORED"   || tempId == "USER" ) {

               // save for next loop
               varId = tempId;
               continue;

            } else if (tempId == "CONSTANT" || tempId == "FINAL")  {
               // process right away
               varId = tempId;

            } else {
               continue;

            }
         }

         // get bool data
         if (varId == "DESIGNABLE" || varId == "SCRIPTABLE" || varId == "STORED" || varId == "USER") {

            // retrieve data until
            varFunc = *index;
            ++index;

            int parenLevel = 0;

            while (true) {

               if (index == temp.end())  {
                  --index;
                  break;
               }

               // pre test
               QString tempId = *index;

               if (tempId == "(") {
                  ++parenLevel;

               } else if (tempId == ")")   {

                  if (parenLevel == 0) {
                     --index;
                     break;

                  } else {
                     --parenLevel;

                  }

               } else if (tempId == "READ"       || tempId == "WRITE"    || tempId == "RESET" ||
                          tempId == "NOTIFY"     || tempId == "REVISION" || tempId == "DESIGNABLE" ||
                          tempId == "SCRIPTABLE" || tempId == "STORED"   || tempId == "USER" ||
                          tempId == "CONSTANT"   || tempId == "FINAL")  {

                  --index;
                  break;
               }

               varFunc = varFunc + " " + tempId;
               ++index;
            }
         }


         if (varId == "READ")  {
            varFunc = *index;
            newWord += preData + "CS_PROPERTY_READ(" + varName + ", " + varFunc  + ")\n";

            varId = "";

            if (isFirst) {
               preData = padding;
               isFirst = false;
            }

         }  else if (varId == "WRITE") {
            varFunc = *index;
            newWord += preData + "CS_PROPERTY_WRITE(" + varName + ", " + varFunc  + ")\n";

            varId = "";

            if (isFirst) {
               preData = padding;
               isFirst = false;
            }

         }  else if (varId == "RESET") {
            varFunc = *index;
            newWord += preData  + "CS_PROPERTY_RESET(" + varName + ", " + varFunc  + ")\n";

            varId = "";

            if (isFirst) {
               preData = padding;
               isFirst = false;
            }

         }  else if (varId == "NOTIFY") {
            varFunc = *index;
            newWord += preData + "CS_PROPERTY_NOTIFY(" + varName + ", " + varFunc + ")\n";

            varId = "";

            if (isFirst) {
               preData = padding;
               isFirst = false;
            }

         }  else if (varId == "REVISION") {
            varFunc = *index;
            newWord += preData + "CS_PROPERTY_REVISION(" + varName + ", " + varFunc + ")\n";

            varId = "";

            if (isFirst) {
               preData = padding;
               isFirst = false;
            }

         }  else if (varId == "DESIGNABLE") {
            // bool
            newWord += preData + "CS_PROPERTY_DESIGNABLE(" + varName + ", " + varFunc + ")\n";

            varId = "";

         }  else if (varId == "SCRIPTABLE") {
            // bool
            newWord += preData + "CS_PROPERTY_SCRIPTABLE(" + varName + ", " + varFunc + ")\n";

            varId = "";

            if (isFirst) {
               preData = padding;
               isFirst = false;
            }

         }  else if (varId == "STORED") {
            // bool
            newWord += preData + "CS_PROPERTY_STORED(" + varName + ", " + varFunc + ")\n";

            varId = "";

            if (isFirst) {
               preData = padding;
               isFirst = false;
            }

         }  else if (varId == "USER") {
            // bool
            newWord += preData + "CS_PROPERTY_USER(" + varName + ", " + varFunc  + ")\n";

            varId = "";

            if (isFirst) {
               preData = padding;
               isFirst = false;
            }

         }  else if (varId == "CONSTANT") {
            // done right away
            newWord += preData + "CS_PROPERTY_CONSTANT(" + varName + ")\n";

            varId = "";

            if (isFirst) {
               preData = padding;
               isFirst = false;
            }

         }  else if (varId == "FINAL") {
            // done right away
            newWord += preData + "CS_PROPERTY_FINAL(" + varName + ")\n";

            varId = "";

            if (isFirst) {
               preData = padding;
               isFirst = false;
            }

         }
      }

      // back out the last line feed
      if (! isFirst) {
         newWord.chop(1);
      }

   } else if (word == "Q_RETURN_ARG")  {
      // ** leave alone


   } else if (word == "Q_SCRIPT_DECLARE_QMETAOBJECT")  {
      // --> change 3  ( unsure )


   } else if (word == "Q_SIGNAL")  {
      // register signal

      // must do this before getting the data!
      QString padding = pmPad(*ptr);

      QString data = pmGetMethod(ptr);
      data = data.trimmed();

      if (! data.isEmpty()) {

         QString methodName;
         QList<QString> paramNames;
         QString signature;

         std::tie(methodName, paramNames, signature) = pmGetSignature(data);

         QString s2 = methodName;

         for (auto index = paramNames.begin(); index < paramNames.end() ; ++index) {
            s2 += ",";
            s2 += *index;
         }

         QString area;

         if (m_section == Section::PUBLIC)  {
            area = "Public";

         } else if (m_section == Section::PROTECTED)  {
            area = "Protected";

         } else if (m_section == Section::PRIVATE)  {
            area = "Private";

         }

         newWord  = "CS_SIGNAL_1(" + area + ", " + signature + ")\n";
         newWord += padding + "CS_SIGNAL_2(" + s2 + ")";
      }


   } else if (word == "Q_SLOT")  {
      // register slot

      // must do this before getting the data!
      QString padding = pmPad(*ptr);

      QString data = pmGetMethod(ptr);
      data = data.trimmed();

      if (! data.isEmpty()) {

         // cs_tag
         bool isCSTag = false;

         int k = data.indexOf(' ');
         QString defineTest = word.left(k);

         if (m_defines.contains(defineTest)) {
            // match
            data    = data.mid(k+1);
            isCSTag = true;
         }

         QString methodName;
         QList<QString> paramNames;
         QString signature;

         std::tie(methodName, paramNames, signature) = pmGetSignature(data);

/*       QByteArray s2 = methodName;

         for (auto index = paramNames.begin(); index < paramNames.end() ; ++index) {
            s2 += ",";
            s2 += *index;
         }
*/

         QString area;

         if (m_section == Section::PUBLIC)  {
            area = "Public";

         } else if (m_section == Section::PROTECTED)  {
            area = "Protected";

         } else if (m_section == Section::PRIVATE)  {
            area = "Private";

         }

         if (isCSTag) {
            // any passed parameters are missing from CS_Tag()
            newWord = "CS_TAG(\"" + defineTest.toLatin1() + "\", " + methodName + "())\n" + padding;

         } else {
            newWord = "";

         }

         newWord += "CS_SLOT_1(" + area + ", " + signature + ")\n";
         newWord += padding + "CS_SLOT_2(" +  methodName + ")";
      }

   } else if (word == "QDOC_PROPERTY")  {
      // ** leave alone, undocumented

   }

   return newWord;
}


// * * *
void csMsg(const QString &msg)
{
   printf("%s\n", qPrintable(msg));
}
